@@ -63,7 +63,7 @@ See L<https://github.com/ukautz/Net-Amazon-DynamoDB> for latest release.
 use Moose;
 
 use v5.10;
-use version 0.74; our $VERSION = qv( "v0.1.10" );
+use version 0.74; our $VERSION = qv( "v0.1.13" );
 
 use Carp qw/ croak /;
 use Data::Dumper;
@@ -74,6 +74,7 @@ use Digest::SHA qw/ sha1_hex sha256_hex sha384_hex sha256 hmac_sha256_base64 /;
 use HTTP::Request;
 use JSON;
 use LWP::UserAgent;
+use LWP::ConnCache;
 use Net::Amazon::AWSSign;
 use Time::HiRes qw/ usleep /;
 use XML::Simple qw/ XMLin /;
@@ -135,13 +136,23 @@ has tables => ( isa => 'HashRef[HashRef]', is => 'rw', required => 1, trigger =>
     }
 } );
 
+=head2 use_keepalives
+
+Use keep_alive connections to AWS (Uses C<LWP::ConnCache> experimental mechanism). 0 to disable, positive number sets value for C<LWP::UserAgent> attribute 'keep_alive'
+Default: 0
+
+=cut
+
+has use_keep_alive => ( isa => 'Int', is => 'rw', default => 0 );
+
 =head2 lwp
 
 Contains C<LWP::UserAgent> instance.
 
 =cut
 
-has lwp => ( isa => 'LWP::UserAgent', is => 'rw', default => sub { LWP::UserAgent->new( timeout => 5 ) } );
+has lwp => ( isa => 'LWP::UserAgent', is => 'rw', lazy => 1, default => sub { my ($self) = @_; LWP::UserAgent->new( timeout => 5, keep_alive => $self->use_keep_alive ) } );
+has _lwpcache => ( isa => 'LWP::ConnCache', is => 'ro', lazy => 1, default => sub { my ($self) = @_; $self->lwp->conn_cache(); } );
 
 =head2 json
 
@@ -185,6 +196,16 @@ Required!
 
 has secret_key => ( isa => 'Str', is => 'rw', required => 1 );
 
+=head2 api_version
+
+AWS API Version. Use format "YYYYMMDD"
+
+Default: 20111205
+
+=cut
+
+has api_version => ( isa => 'Str', is => 'rw', default => '20111205' );
+
 =head2 read_consistent
 
 Whether reads (get_item) consistent per default or not. This does not affcect batch_get_item or scan_items or query_items, which are always eventually consistent.
@@ -467,6 +488,7 @@ sub describe_table {
     # got result
     if ( $res_ok ) {
         if ( defined $json_ref->{ Table } ) {
+            no warnings 'uninitialized';
             return {
                 existing      => 1,
                 size          => $json_ref->{ Table }->{ TableSizeBytes },
@@ -650,7 +672,8 @@ sub put_item {
     $args_ref ||= {
         return_old  => 0,
         no_cache    => 0,
-        use_cache   => 0
+        use_cache   => 0,
+        max_retries => undef
     };
     $table = $self->_table_name( $table );
     
@@ -697,7 +720,9 @@ sub put_item {
     $put{ ReturnValues } = 'ALL_OLD' if $args_ref->{ return_old };
     
     # perform create
-    my ( $res, $res_ok, $json_ref ) = $self->request( PutItem => \%put );
+    my ( $res, $res_ok, $json_ref ) = $self->request( PutItem => \%put, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     # get result
     if ( $res_ok ) {
@@ -724,6 +749,173 @@ sub put_item {
 }
 
 
+=head2 batch_write_item $tables_ref, [$args_ref]
+
+Batch put / delete items into one ore more tables.
+
+Caution: Each batch put / delete cannot process more operations than you have write capacity for the table.
+
+Example:
+
+    my ( $ok, $unprocessed_count, $next_query_ref ) = $ddb->batch_write_item( {
+        table_name => {
+            put => [
+                {
+                    attrib1 => "Value 1",
+                    attrib2 => "Value 2",
+                },
+                # { .. } ..
+            ],
+            delete => [
+                {
+                    hash_key => "Hash Key Value",
+                    range_key => "Range Key Value",
+                },
+                # { .. } ..
+            ]
+        },
+        # table2_name => ..
+    } );
+    
+    if ( $ok ) {
+        if ( $unprocessed_count ) {
+            print "Ok, but $unprocessed_count still not processed\n";
+            $ddb->batch_write_item( $next_query_ref );
+        }
+        else {
+            print "All processed\n";
+        }
+    }
+
+=over
+
+=item $tables_ref
+
+HashRef in the form
+
+    { table_name => { put => [ { attribs }, .. ], delete => [ { primary keys } ] } }
+
+=item $args_ref
+
+HashRef
+
+=over
+
+=item * process_all
+
+Keep processing everything which is returned as unprocessed (if you send more operations than your
+table has write capability or you surpass the max amount of operations OR max size of request (see AWS API docu)).
+
+Caution: Error handling
+
+Default: 0
+
+=back
+
+=back
+
+=cut
+
+sub batch_write_item {
+    my ( $self, $tables_ref, $args_ref ) = @_;
+    $args_ref ||= {
+        process_all => 0,
+        max_retries => undef
+    };
+    
+    # check definition
+    my %table_map;
+    foreach my $table( keys %$tables_ref ) {
+        $table = $self->_table_name( $table );
+        my $table_ref = $self->_check_table( "batch_write_item", $table );
+        $table_map{ $table } = $table_ref;
+    }
+    
+    my %write = ( RequestItems => {} );
+    foreach my $table( keys %table_map ) {
+        my $table_out = $self->_table_name( $table, 1 );
+        my $t_ref = $tables_ref->{ $table_out };
+        my $table_requests_ref = $write{ RequestItems }->{ $table } = [];
+        
+        foreach my $operation( qw/ put delete / ) {
+            next unless defined $t_ref->{ $operation };
+            my @operations = ref( $t_ref->{ $operation } ) eq 'ARRAY'
+                ? @{ $t_ref->{ $operation } }
+                : ( $t_ref->{ $operation } );
+            
+            # put ..
+            if ( $operation eq 'put' ) {
+                foreach my $put_ref( @operations ) {
+                    push @$table_requests_ref, { 'PutRequest' => { Item => my $request_ref = {} } };
+                    
+                    # build the item
+                    foreach my $key( keys %$put_ref ){
+                        my $type = $self->_attrib_type( $table, $key );
+                        my $value;
+                        if ( $type eq 'SS' || $type eq 'NS' ) {
+                            my @values = map { $_. '' } ( ref( $put_ref->{ $key } ) ? @{ $put_ref->{ $key } } : () );
+                            $value = \@values;
+                        }
+                        else {
+                            $value = $put_ref->{ $key } .'';
+                        }
+                        $request_ref->{ $key } = { $type => $value };
+                    }
+                }
+            }
+            
+            # delete ..
+            else {
+                foreach my $delete_ref( @operations ) {
+                    push @$table_requests_ref, { 'DeleteRequest' => { Key => my $request_ref = {} } };
+                    $self->_build_pk_filter( $table, $delete_ref, $request_ref );
+                }
+            }
+        }
+    }
+    
+    # perform create
+    my ( $res, $res_ok, $json_ref ) = $self->request( BatchWriteItem => \%write, {
+        max_retries => $args_ref->{ max_retries },
+    } );
+    
+    # having more to process
+    while ( $args_ref->{ process_all }
+        && $res_ok
+        && defined $json_ref->{ UnprocessedItems }
+        && scalar( keys %{ $json_ref->{ UnprocessedItems } } )
+    ) {
+        ( $res, $res_ok, $json_ref ) = $self->request( BatchWriteItem => {
+            RequestItems => $json_ref->{ UnprocessedItems }
+        }, {
+            max_retries => $args_ref->{ max_retries },
+        } );
+    }
+    
+    # count unprocessed
+    my $unprocessed_count = 0;
+    my %next_query;
+    if ( $res_ok && defined $json_ref->{ UnprocessedItems } ) {
+        foreach my $table( keys %{ $json_ref->{ UnprocessedItems } } ) {
+            my @operations = @{ $json_ref->{ UnprocessedItems }->{ $table } };
+            next unless @operations;
+            $unprocessed_count += scalar( @operations );
+            $next_query{ $table } = {};
+            foreach my $operation_ref( @operations ) {
+                my ( $item_ref, $operation_name ) = defined $operation_ref->{ PutRequest }
+                    ? ( $operation_ref->{ PutRequest }->{ Item }, 'put' )
+                    : ( $operation_ref->{ DeleteRequest }->{ Key }, 'delete' );
+                #print Dumper( [ $operation_ref, $operation_name, $item_ref ] );
+                push @{ $next_query{ $table }->{ $operation_name } ||= [] },
+                    $self->_format_item( $table, $item_ref )
+            }
+        }
+    }
+    
+    return wantarray ? ( $res_ok, $unprocessed_count, \%next_query ) : $res_ok;
+}
+
+
 
 =head2 update_item $table, $update_ref, $where_ref, [$args_ref]
 
@@ -812,12 +1004,13 @@ sub update_item {
     $args_ref ||= {
         return_mode => '',
         no_cache    => 0,
-        use_cache   => 0
+        use_cache   => 0,
+        max_retries => undef
     };
     $table = $self->_table_name( $table );
     
     # check definition
-    my $table_ref = $self->_check_table( "put_item", $table );
+    my $table_ref = $self->_check_table( "update_item", $table );
     
     croak "update_item: Cannot update hash key value, do not set it in update-clause"
         if defined $update_ref->{ $table_ref->{ hash_key } };
@@ -837,14 +1030,14 @@ sub update_item {
         );
     
     # check other attributes
-    $self->_check_keys( "put_item: item values", $table, $update_ref );
+    $self->_check_keys( "update_item: item values", $table, $update_ref );
     croak "update_item: Cannot update hash key '$table_ref->{ hash_key }'. You have to delete and put the item!"
         if defined $update_ref->{ $table_ref->{ hash_key } };
     croak "update_item: Cannot update range key '$table_ref->{ hash_key }'. You have to delete and put the item!"
         if defined $table_ref->{ range_key } && defined $update_ref->{ $table_ref->{ range_key } };
     
     # having where -> check now
-    $self->_check_keys( "put_item: where clause", $table, $where_ref );
+    $self->_check_keys( "update_item: where clause", $table, $where_ref );
     
     # build put
     my %update = (
@@ -913,7 +1106,9 @@ sub update_item {
     }
     
     # perform create
-    my ( $res, $res_ok, $json_ref ) = $self->request( UpdateItem => \%update );
+    my ( $res, $res_ok, $json_ref ) = $self->request( UpdateItem => \%update, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     # get result
     if ( $res_ok ) {
@@ -1013,10 +1208,11 @@ sub get_item {
     my ( $self, $table, $pk_ref, $args_ref ) = @_;
     $table = $self->_table_name( $table );
     $args_ref ||= {
-        consistent => undef,
-        attributes => undef,
-        no_cache   => 0,
-        use_cache  => 0
+        consistent  => undef,
+        attributes  => undef,
+        no_cache    => 0,
+        use_cache   => 0,
+        max_retries => undef
     };
     $args_ref->{ consistent } //= $self->read_consistent;
     
@@ -1065,7 +1261,9 @@ sub get_item {
     }
     
     # perform create
-    my ( $res, $res_ok, $json_ref ) = $self->request( GetItem => \%get );
+    my ( $res, $res_ok, $json_ref ) = $self->request( GetItem => \%get, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     # return on success
     my $item_ref = $self->_format_item( $table, $json_ref->{ Item } ) if $res_ok && defined $json_ref->{ Item };
@@ -1123,7 +1321,9 @@ HashRef of tablename => primary key ArrayRef
 
 sub batch_get_item {
     my ( $self, $tables_ref, $args_ref ) = @_;
-    $args_ref ||= {};
+    $args_ref ||= {
+        max_retries => undef
+    };
     
     
     # check definition
@@ -1177,7 +1377,9 @@ sub batch_get_item {
     }
     
     # perform create
-    my ( $res, $res_ok, $json_ref ) = $self->request( BatchGetItem => \%get );
+    my ( $res, $res_ok, $json_ref ) = $self->request( BatchGetItem => \%get, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     # return on success
     if ( $res_ok && defined $json_ref->{ Responses } ) {
@@ -1251,9 +1453,10 @@ Force using cache, if disabled per default but setupped
 sub delete_item {
     my ( $self, $table, $where_ref, $args_ref ) = @_;
     $args_ref ||= {
-        return_old => 0,
-        no_cache   => 0,
-        use_cache  => 0
+        return_old  => 0,
+        no_cache    => 0,
+        use_cache   => 0,
+        max_retries => undef
     };
     $table = $self->_table_name( $table );
     
@@ -1303,7 +1506,9 @@ sub delete_item {
     }
     
     # perform create
-    my ( $res, $res_ok, $json_ref ) = $self->request( DeleteItem => \%delete );
+    my ( $res, $res_ok, $json_ref ) = $self->request( DeleteItem => \%delete, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     if ( $res_ok ) {
         
@@ -1459,6 +1664,7 @@ sub query_items {
         attributes  => undef,   # eq [ qw/ attrib1 attrib2 / ]
         count       => 0,       # returns amount instead of the actual result
         all         => 0,       # read all entries (runs possibly multiple queries)
+        max_retries => undef,   # overwrite default max rewrites
     };
     
     # check definition
@@ -1543,7 +1749,9 @@ sub query_items {
     
     # perform query
     #print Dumper( { QUERY => \%query } );
-    my ( $res, $res_ok, $json_ref ) = $self->request( Query => \%query );
+    my ( $res, $res_ok, $json_ref ) = $self->request( Query => \%query, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     # format & return result
     if ( $res_ok && defined $json_ref->{ Items } ) {
@@ -1626,6 +1834,7 @@ sub scan_items {
         attributes  => undef,   # eq [ qw/ attrib1 attrib2 / ]
         count       => 0,       # returns amount instead of the actual result
         all         => 0,       # read all entries (runs possibly multiple queries)
+        max_retries => undef,   # overwrite default max retries
     };
     
     # check definition
@@ -1700,7 +1909,9 @@ sub scan_items {
     }
     
     # perform query
-    my ( $res, $res_ok, $json_ref ) = $self->request( Scan => \%query );
+    my ( $res, $res_ok, $json_ref ) = $self->request( Scan => \%query, {
+        max_retries => $args_ref->{ max_retries },
+    } );
     
     # format & return result
     if ( $res_ok && defined $json_ref->{ Items } ) {
@@ -1771,7 +1982,10 @@ Arbitrary request to DynamoDB API
 =cut
 
 sub request {
-    my ( $self, $target, $json ) = @_;
+    my ( $self, $target, $json, $args_ref ) = @_;
+    $args_ref ||= {
+        max_retries => undef
+    };
     
     # assure security token existing
     unless( $self->_init_security_token() ) {
@@ -1791,7 +2005,7 @@ sub request {
         'host:'. $self->host,
         'x-amz-date:'. $http_date,
         'x-amz-security-token:'. $self->_credentials->{ SessionToken },
-        'x-amz-target:DynamoDB_20111205.'. $target,
+        'x-amz-target:DynamoDB_'. $self->api_version. '.'. $target,
         '',
         $json
     );
@@ -1804,7 +2018,7 @@ sub request {
     # .. setup headers
     $request->header( host => $self->host );
     $request->header( 'x-amz-date' => $http_date );
-    $request->header( 'x-amz-target', 'DynamoDB_20111205.'. $target );
+    $request->header( 'x-amz-target', 'DynamoDB_'. $self->api_version. '.'. $target );
     $request->header( 'x-amzn-authorization' => join( ',',
         'AWS3 AWSAccessKeyId='. $self->_credentials->{ AccessKeyId },
         'Algorithm=HmacSHA256',
@@ -1818,12 +2032,15 @@ sub request {
     $request->content( $json );
     
     my ( $json_ref, $response );
-    my $tries = $self->max_retries + 1;
+    my $tries = defined $args_ref->{ max_retries }
+        ? $args_ref->{ max_retries }
+        : $self->max_retries + 1;
     while( 1 ) {
         
         # run request
         $response = $self->lwp->request( $request );
         $ENV{ DYNAMO_DB_DEBUG } && warn Dumper( $response );
+        $ENV{ DYNAMO_DB_DEBUG_KEEPALIVE } && warn "  LWP keepalives in use: ", scalar($self->_lwpcache()->get_connections()), "/", $self->_lwpcache()->total_capacity(), "\n";
         
         # get json
         $json_ref = $response
@@ -2082,9 +2299,20 @@ sub _format_item {
     my ( $self, $table, $from_ref ) = @_;
     my $table_ref = $self->_check_table( format_item => $table );
     my %formatted;
-    while( my( $attrib, $type ) = each %{ $table_ref->{ attributes } } ) {
-        next unless defined $from_ref->{ $attrib };
-        $formatted{ $attrib } = $from_ref->{ $attrib }->{ $type };
+    if ( defined $from_ref->{ HashKeyElement } ) {
+        my @keys = ( 'hash' );
+        push @keys, 'range' if defined $table_ref->{ range_key };
+        foreach my $key( @keys ) {
+            my $key_name = $table_ref->{ "${key}_key" };
+            my $key_type = $table_ref->{ attributes }->{ $key_name };
+            $formatted{ $key_name } = $from_ref->{ ucfirst( $key ). 'KeyElement' }->{ $key_type };
+        }
+    }
+    else {
+        while( my( $attrib, $type ) = each %{ $table_ref->{ attributes } } ) {
+            next unless defined $from_ref->{ $attrib };
+            $formatted{ $attrib } = $from_ref->{ $attrib }->{ $type };
+        }
     }
     return \%formatted;
 }
@@ -2169,6 +2397,8 @@ __PACKAGE__->meta->make_immutable;
 
 =item * Thanks to MadHacker L<http://stackoverflow.com/users/1139526/madhacker> (the signing code in request method)
 
+=item * Benjamin Abbott-Scoot <benjamin@abbott-scott.net> (Keep Alive patch)
+
 =back
 
 =head1 COPYRIGHT
@@ -2181,4 +2411,4 @@ Same license as Perl itself.
 
 =cut
 
-1;
\ No newline at end of file
+1;
